/**
 * @defgroup framework Mock Turtle Firmware Framework
 * @{
 * @copyright (C) 2015-2016 CERN (www.cern.ch)
 * @author Federico Vaga <federico.vaga@cern.ch>
 *
 * SPDX-License-Identifier: LGPL-3.0-or-later
 */

#ifndef __TRTL_FW_FRM_H__
#define __TRTL_FW_FRM_H__

#include <inttypes.h>
#include <generated/autoconf.h>
#include <hw/mockturtle_config.h>
#include "mockturtle-rt.h"


/**
 * Action prototype type
 * @param[in] msg_i input message
 * @param[out] msg_o output message
 * @return 0 on success. -1 on error
 *
 * The header for the output message is prepared by the framework.
 * The header's fields that the user should touch are
 *
 * On error the message will be sent anyway to the host. This
 * is just in case of future development.
 */
typedef int (action_t)(struct trtl_msg *msg_i, struct trtl_msg *msg_o);

/**
 * Variable flag. Register
 */
#define TRTL_RT_VARIABLE_FLAG_FLD (1 << 0)

/**
 * Description of a variable that we want to export to the external
 * world (host or network).
 */
struct rt_variable {
	void *addr; /**< variable address */
	uint32_t mask; /**< variable mask without offset applied */
	uint8_t offset; /**< variable offset within the word */
	uint32_t flags; /**< variable options */
};


/**
 * Description of a buffer that we want to export to the external world
 * (host or network)
 */
struct rt_buffer {
	void *buf; /**< structure location */
	uint32_t len; /**< data structure lenght */
	/* Maybe other option later in time */
};


/**
 * Real-Time Application Descriptor
 */
struct rt_application {
	const char name[16]; /**< Firmware name*/
	const uint32_t *fpga_id_compat; /**< list of compatible FPGA
					   application ID */
	const unsigned int fpga_id_compat_n; /**< number of entry in
						the fpga_id_compat list */
	const struct trtl_rt_version version; /**< version running */

	struct rt_buffer *buffers; /**< exported buffers */
	unsigned int n_buffers; /**< number or exported buffers */

	struct rt_variable *variables; /**< exported variables */
	unsigned int n_variables; /**< number or exported variables */

	action_t **actions; /**< list of custum actions */
	unsigned int n_actions; /**< number of custum actions */

	unsigned int seq; /** sequence number reference */
	const struct trtl_config_rom *cfgrom;
};


extern struct rt_application *_app;

extern int rt_init(struct rt_application *app);
extern void rt_get_time(uint32_t *seconds, uint32_t *cycles);
extern void rt_print_header(struct trtl_hmq_header *h);
extern void rt_print_data(uint32_t *d, unsigned int count);
extern void rt_print_message(struct trtl_msg *msg);
extern int trtl_rt_mq_send_uint32(enum trtl_mq_type type,
				  unsigned int idx_mq,
				  uint8_t msg_id,
				  unsigned int n,
				  ...);
extern int trtl_rt_mq_send_buf(enum trtl_mq_type type,
			       unsigned int idx_mq,
			       uint8_t msg_id,
			       unsigned int n,
			       void *data);
extern int rt_mq_action_dispatch(enum trtl_mq_type type,
				 unsigned int idx_mq);
extern void rt_message_error(struct trtl_msg *msg, int err);

#endif
/**@}*/
