/**
 * @defgroup library-message-queue Message Queue
 * @{
 * @ingroup library
 * @brief Message Queues definitions and functions
 * @copyright (C) 2013-2016 CERN (www.cern.ch)
 * @author Tomasz Wlostowski <tomasz.wlostowski@cern.ch>
 * @author Federico Vaga <federico.vaga@cern.ch>
 *
 * SPDX-License-Identifier: LGPL-3.0-or-later
 */

#ifndef __TRTL_RT_MQUEUE_H
#define __TRTL_RT_MQUEUE_H

#include <hw/mockturtle_addresses.h>
#include <hw/mockturtle_queue.h>
#include <hw/mockturtle_cpu_lr.h>

#include "mockturtle-rt-common.h"


#define TRTL_MQ_SLOT_IN(slot)  (TRTL_MQ_BASE_IN + ((slot) << 16))
// Outgoung slot base address, relative to BASE_HMQ
#define TRTL_MQ_SLOT_OUT(slot) (TRTL_MQ_BASE_OUT + ((slot) << 16))
/**
 * List of Message Queue types
 */
enum trtl_mq_type {
	TRTL_HMQ = 0, /**< Host Message Queue - Host-Firmware communication */
	TRTL_RMQ, /**< Remote Message Queue - Network-Firmware communication */
	__TRTL_MAX_MQ_TYPE,
};


/**
 * It gets the Message Queue base address
 * @param[in] type MQ type
 */
static inline void *trtl_mq_base_address(enum trtl_mq_type type)
{
	return (void *)(type == TRTL_HMQ ? TRTL_ADDR_HMQ_BASE : TRTL_ADDR_RMQ_BASE);
}


/**
 * It writes on a Message Queue register
 * @param[in] type MQ type to use
 * @param[val] value to write
 * @param[reg] reg register offset
 */
static inline void mq_writel(enum trtl_mq_type type, uint32_t val, uint32_t reg)
{
	*(volatile uint32_t *)(trtl_mq_base_address(type) + reg) = val;
}


/**
 * @copydoc TRTL_MQ_CMD_CLAIM
 * @param[in] type MQ type to use
 * @param[in] slot slot number
 */
static inline void mq_claim(enum trtl_mq_type type, int slot)
{
	mq_writel(type, TRTL_MQ_CMD_CLAIM, TRTL_MQ_SLOT_OUT(slot) + TRTL_MQ_SLOT_COMMAND);
}

/**
 * @copydoc TRTL_MQ_CMD_PURGE
 * @param[in] type MQ type to use
 * @param[in] slot slot number
 */
static inline void mq_purge(enum trtl_mq_type type, int slot)
{
	mq_writel(type, TRTL_MQ_CMD_PURGE, TRTL_MQ_SLOT_OUT(slot) + TRTL_MQ_SLOT_COMMAND);
	mq_writel(type, TRTL_MQ_CMD_PURGE, TRTL_MQ_SLOT_IN(slot) + TRTL_MQ_SLOT_COMMAND);

}


/**
 * @copydoc TRTL_MQ_CMD_READY
 * @param[in] type MQ type to use
 * @param[in] slot slot number
 */
static inline void mq_send(enum trtl_mq_type type, int slot)
{
	mq_writel(type, TRTL_MQ_CMD_READY, TRTL_MQ_SLOT_OUT(slot) + TRTL_MQ_SLOT_COMMAND);
}


/**
 * @copydoc TRTL_MQ_CMD_DISCARD
 * @param[in] type MQ type to use
 * @param[in] slot slot number
 */
static inline void mq_discard(enum trtl_mq_type type, int slot)
{
	mq_writel(type, TRTL_MQ_CMD_DISCARD, TRTL_MQ_SLOT_IN(slot));
}


/**
 * It gets the output slot data field pointer
 * @param[in] type MQ type to use
 * @param[in] slot slot number
 */
static inline void *mq_map_out_buffer(enum trtl_mq_type type, int slot)
{
	return (void *) (trtl_mq_base_address(type) +
			 TRTL_MQ_SLOT_OUT (slot) + TRTL_MQ_SLOT_DATA_START);
}


/**
 * It gets the input slot data field pointer
 * @param[in] type MQ type to use
 * @param[in] slot slot number
 */
static inline void *mq_map_in_buffer(enum trtl_mq_type type, int slot)
{
	return (void *) (trtl_mq_base_address(type) +
			 TRTL_MQ_SLOT_IN (slot) + TRTL_MQ_SLOT_DATA_START);
}


/**
 * It gets the output slot header field pointer
 * @param[in] type MQ type to use
 * @param[in] slot slot number
 */
static inline void *mq_map_out_header(enum trtl_mq_type type, int slot)
{
	return (void *) (trtl_mq_base_address(type) +
			 TRTL_MQ_SLOT_OUT (slot) + TRTL_MQ_SLOT_HEADER_START);
}


/**
 * It gets the input slot header field pointer
 * @param[in] type MQ type to use
 * @param[in] slot slot number
 */
static inline void *mq_map_in_header(enum trtl_mq_type type, int slot)
{
	return (void *) (trtl_mq_base_address(type) +
			 TRTL_MQ_SLOT_IN (slot) + TRTL_MQ_SLOT_HEADER_START);
}

/**
 * It maps a given MQ for outcoming messages
 * @param[in] type MQ type to use
 * @param[in] idx_mq MQ index
 * @param[out] msg where to map the message
 */
static inline void mq_map_out_message(enum trtl_mq_type type,
				     unsigned idx_mq,
				     struct trtl_msg *msg)
{
	msg->header = mq_map_out_header(type, idx_mq);
	msg->payload = mq_map_out_buffer(type, idx_mq);
}

/**
 * It maps a given MQ for incoming messages
 * @param[in] type MQ type to use
 * @param[in] idx_mq MQ index
 * @param[out] msg where to map the message
 */
static inline void mq_map_in_message(enum trtl_mq_type type,
				     unsigned idx_mq,
				     struct trtl_msg *msg)
{
	msg->header = mq_map_in_header(type, idx_mq);
	msg->payload = mq_map_in_buffer(type, idx_mq);
}


/**
 * It gets the current MQ input status
 * @return message queues input status bitmask
 */
static inline uint32_t mq_poll_in(enum trtl_mq_type type)
{
	uint32_t poll = lr_readl(MT_CPU_LR_REG_HMQ_STAT + (type * 4));

	/* HMQ and RMQ have the same format -> use the same mask */
	return ((poll & MT_CPU_LR_HMQ_STAT_IN_MASK) >> MT_CPU_LR_HMQ_STAT_IN_SHIFT);
}

/**
 * It gets the current MQ output status
 * @return message queues output status bitmask
 */
static inline uint32_t mq_poll_out(enum trtl_mq_type type)
{
	uint32_t poll = lr_readl(MT_CPU_LR_REG_HMQ_STAT + (type * 4));

	/* HMQ and RMQ have the same format -> use the same mask */
	return ((poll & MT_CPU_LR_HMQ_STAT_OUT_MASK) >> MT_CPU_LR_HMQ_STAT_OUT_SHIFT);
}

#endif
/**@}*/
